@@ -38,10 +38,13 @@ extern "C" {
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <signal.h>
+#include <unistd.h>
 }
 
 #include <cstdlib>
 #include <fstream>
+#include <iomanip>
 #include <iostream>
 #include <stdexcept>
 
@@ -63,14 +66,105 @@ safe_mkdir(const char* path)
 
 static
 void
-touch(const char* path)
+touch(const std::string& path)
 {
-    std::ofstream os(path);
+    std::ofstream os(path.c_str());
     if (!os)
-        ATF_FAIL(std::string("Could not create file ") + path);
+        ATF_FAIL("Could not create file " + path);
     os.close();
 }
 
+// ------------------------------------------------------------------------
+// Helper tests for "t_cleanup".
+// ------------------------------------------------------------------------
+
+ATF_TEST_CASE_WITH_CLEANUP(cleanup_pass);
+ATF_TEST_CASE_HEAD(cleanup_pass)
+{
+    set("descr", "Helper test case for the t_cleanup test program");
+}
+ATF_TEST_CASE_BODY(cleanup_pass)
+{
+    touch(config().get("tmpfile"));
+}
+ATF_TEST_CASE_CLEANUP(cleanup_pass)
+{
+    if (config().get_bool("cleanup"))
+        atf::fs::remove(atf::fs::path(config().get("tmpfile")));
+}
+
+ATF_TEST_CASE_WITH_CLEANUP(cleanup_fail);
+ATF_TEST_CASE_HEAD(cleanup_fail)
+{
+    set("descr", "Helper test case for the t_cleanup test program");
+}
+ATF_TEST_CASE_BODY(cleanup_fail)
+{
+    touch(config().get("tmpfile"));
+    ATF_FAIL("On purpose");
+}
+ATF_TEST_CASE_CLEANUP(cleanup_fail)
+{
+    if (config().get_bool("cleanup"))
+        atf::fs::remove(atf::fs::path(config().get("tmpfile")));
+}
+
+ATF_TEST_CASE_WITH_CLEANUP(cleanup_skip);
+ATF_TEST_CASE_HEAD(cleanup_skip)
+{
+    set("descr", "Helper test case for the t_cleanup test program");
+}
+ATF_TEST_CASE_BODY(cleanup_skip)
+{
+    touch(config().get("tmpfile"));
+    ATF_SKIP("On purpose");
+}
+ATF_TEST_CASE_CLEANUP(cleanup_skip)
+{
+    if (config().get_bool("cleanup"))
+        atf::fs::remove(atf::fs::path(config().get("tmpfile")));
+}
+
+ATF_TEST_CASE_WITH_CLEANUP(cleanup_curdir);
+ATF_TEST_CASE_HEAD(cleanup_curdir)
+{
+    set("descr", "Helper test case for the t_cleanup test program");
+}
+ATF_TEST_CASE_BODY(cleanup_curdir)
+{
+    std::ofstream os("oldvalue");
+    if (!os)
+        ATF_FAIL("Failed to create oldvalue file");
+    os << 1234;
+    os.close();
+}
+ATF_TEST_CASE_CLEANUP(cleanup_curdir)
+{
+    std::ifstream is("oldvalue");
+    if (is) {
+        int i;
+        is >> i;
+        std::cout << "Old value: " << i << std::endl;
+        is.close();
+    }
+}
+
+ATF_TEST_CASE_WITH_CLEANUP(cleanup_sigterm);
+ATF_TEST_CASE_HEAD(cleanup_sigterm)
+{
+    set("descr", "Helper test case for the t_cleanup test program");
+}
+ATF_TEST_CASE_BODY(cleanup_sigterm)
+{
+    touch(config().get("tmpfile"));
+    ::kill(::getpid(), SIGTERM);
+    touch(config().get("tmpfile") + ".no");
+}
+ATF_TEST_CASE_CLEANUP(cleanup_sigterm)
+{
+    atf::fs::remove(atf::fs::path(config().get("tmpfile")));
+}
+
 // ------------------------------------------------------------------------
 // Helper tests for "t_config".
 // ------------------------------------------------------------------------
@@ -172,6 +266,19 @@ ATF_TEST_CASE_BODY(fork_mangle_fds)
 #endif
 }
 
+ATF_TEST_CASE(fork_umask);
+ATF_TEST_CASE_HEAD(fork_umask)
+{
+    set("descr", "Helper test case for the t_fork test program");
+}
+ATF_TEST_CASE_BODY(fork_umask)
+{
+    mode_t m = ::umask(0);
+    std::cout << "umask: " << std::setw(4) << std::setfill('0')
+              << std::oct << m << std::endl;
+    (void)::umask(m);
+}
+
 // ------------------------------------------------------------------------
 // Helper tests for "t_ident".
 // ------------------------------------------------------------------------
@@ -343,6 +450,13 @@ ATF_TEST_CASE_BODY(srcdir_exists)
 
 ATF_INIT_TEST_CASES(tcs)
 {
+    // Add helper tests for t_cleanup.
+    tcs.push_back(&cleanup_pass);
+    tcs.push_back(&cleanup_fail);
+    tcs.push_back(&cleanup_skip);
+    tcs.push_back(&cleanup_curdir);
+    tcs.push_back(&cleanup_sigterm);
+
     // Add helper tests for t_config.
     tcs.push_back(&config_unset);
     tcs.push_back(&config_empty);
@@ -355,6 +469,7 @@ ATF_INIT_TEST_CASES(tcs)
 
     // Add helper tests for t_fork.
     tcs.push_back(&fork_mangle_fds);
+    tcs.push_back(&fork_umask);
 
     // Add helper tests for t_meta_data.
     tcs.push_back(&ident_1);
