@@ -42,6 +42,7 @@
 #include <vector>
 
 #include "atf/application.hpp"
+#include "atf/fs.hpp"
 #include "atf/formats.hpp"
 #include "atf/text.hpp"
 #include "atf/ui.hpp"
@@ -82,8 +83,8 @@ public:
     virtual ~writer(void) {}
 
     virtual void write_ntps(size_t) {}
-    virtual void write_tp_start(const atf::fs::path&, size_t) {}
-    virtual void write_tp_end(void) {}
+    virtual void write_tp_start(const std::string&, size_t) {}
+    virtual void write_tp_end(const std::string&) {}
     virtual void write_tc_start(const std::string&) {}
     virtual void write_tc_stdout_line(const std::string&) {}
     virtual void write_tc_stderr_line(const std::string&) {}
@@ -117,9 +118,16 @@ public:
 
     virtual
     void
-    write_tp_start(const atf::fs::path& name, size_t ntcs)
+    write_tp_start(const std::string& name, size_t ntcs)
     {
-        m_tpname = name.str();
+        m_tpname = name;
+    }
+
+    virtual
+    void
+    write_tp_end(const std::string& reason)
+    {
+        // TODO: Report failed test programs in some way.
     }
 
     virtual
@@ -166,7 +174,8 @@ class ticker_writer : public writer {
     size_t m_curtp, m_ntps;
     size_t m_tcs_passed, m_tcs_failed, m_tcs_skipped;
     std::string m_tcname, m_tpname;
-    std::vector< std::string > m_failed;
+    std::vector< std::string > m_failed_tcs;
+    std::vector< std::string > m_failed_tps;
 
     void
     write_ntps(size_t ntps)
@@ -179,14 +188,14 @@ class ticker_writer : public writer {
     }
 
     void
-    write_tp_start(const atf::fs::path& tp, size_t ntcs)
+    write_tp_start(const std::string& tp, size_t ntcs)
     {
         using atf::text::to_string;
         using atf::ui::format_text;
 
-        m_tpname = tp.str();
+        m_tpname = tp;
 
-        (*m_os) << format_text(tp.str() + " (" + to_string(m_curtp) +
+        (*m_os) << format_text(tp + " (" + to_string(m_curtp) +
                                "/" + to_string(m_ntps) + "): " +
                                to_string(ntcs) + " test cases")
                 << std::endl;
@@ -194,12 +203,24 @@ class ticker_writer : public writer {
     }
 
     void
-    write_tp_end(void)
+    write_tp_end(const std::string& reason)
     {
+        using atf::ui::format_text_with_tag;
+
         m_curtp++;
 
+        if (!reason.empty()) {
+            (*m_os) << format_text_with_tag("BOGUS TEST PROGRAM: Cannot "
+                                            "trust its results because "
+                                            "of `" + reason + "'",
+                                            m_tpname + ": ", false)
+                    << std::endl;
+            m_failed_tps.push_back(m_tpname);
+        }
         (*m_os) << std::endl;
         (*m_os).flush();
+
+        m_tpname.clear();
     }
 
     void
@@ -223,7 +244,7 @@ class ticker_writer : public writer {
         } else if (s == atf::tests::tcr::status_failed) {
             str = "Failed: " + tcr.get_reason();
             m_tcs_failed++;
-            m_failed.push_back(m_tpname + ":" + m_tcname);
+            m_failed_tcs.push_back(m_tpname + ":" + m_tcname);
         } else if (s == atf::tests::tcr::status_skipped) {
             str = "Skipped: " + tcr.get_reason();
             m_tcs_skipped++;
@@ -234,6 +255,8 @@ class ticker_writer : public writer {
         // to specify the current column, which is needed because we have
         // already printed the tc's name.
         (*m_os) << str << std::endl;
+
+        m_tcname = "";
     }
 
     void
@@ -244,9 +267,17 @@ class ticker_writer : public writer {
         using atf::ui::format_text;
         using atf::ui::format_text_with_tag;
 
-        if (!m_failed.empty()) {
+        if (!m_failed_tps.empty()) {
+            (*m_os) << format_text("Failed (bogus) test programs:")
+                    << std::endl;
+            (*m_os) << format_text_with_tag(join(m_failed_tps, ", "),
+                                            "    ", false) << std::endl
+                    << std::endl;
+        }
+
+        if (!m_failed_tcs.empty()) {
             (*m_os) << format_text("Failed test cases:") << std::endl;
-            (*m_os) << format_text_with_tag(join(m_failed, ", "),
+            (*m_os) << format_text_with_tag(join(m_failed_tcs, ", "),
                                             "    ", false) << std::endl
                     << std::endl;
         }
@@ -299,7 +330,7 @@ class converter : public atf::formats::atf_tps_reader {
     }
 
     void
-    got_tp_start(const atf::fs::path& tp, size_t ntcs)
+    got_tp_start(const std::string& tp, size_t ntcs)
     {
         for (outs_vector::iterator iter = m_outs.begin();
              iter != m_outs.end(); iter++)
@@ -307,11 +338,11 @@ class converter : public atf::formats::atf_tps_reader {
     }
 
     void
-    got_tp_end(void)
+    got_tp_end(const std::string& reason)
     {
         for (outs_vector::iterator iter = m_outs.begin();
              iter != m_outs.end(); iter++)
-            (*iter)->write_tp_end();
+            (*iter)->write_tp_end(reason);
     }
 
     void
