@@ -56,6 +56,7 @@ extern "C" {
 #include "atf/formats.hpp"
 #include "atf/fs.hpp"
 #include "atf/io.hpp"
+#include "atf/parser.hpp"
 #include "atf/tests.hpp"
 #include "atf/text.hpp"
 
@@ -86,6 +87,7 @@ class muxer : public atf::formats::atf_tcs_reader {
     atf::fs::path m_tp;
     atf::formats::atf_tps_writer m_writer;
 
+    bool m_finalized;
     size_t m_ntcs;
     std::string m_tcname;
 
@@ -95,7 +97,10 @@ class muxer : public atf::formats::atf_tcs_reader {
     void
     got_ntcs(size_t ntcs)
     {
-        m_writer.start_tp(m_tp, ntcs);
+        m_writer.start_tp(m_tp.str(), ntcs);
+        if (ntcs == 0)
+            throw atf::serial::format_error("Bogus test program: reported "
+                                            "0 test cases");
     }
 
     void
@@ -120,6 +125,7 @@ class muxer : public atf::formats::atf_tcs_reader {
         }
 
         m_writer.end_tc(tcr);
+        m_tcname = "";
     }
 
     void
@@ -140,15 +146,36 @@ public:
         atf::formats::atf_tcs_reader(is),
         m_tp(tp),
         m_writer(w),
+        m_finalized(false),
         m_passed(0),
         m_failed(0),
         m_skipped(0)
     {
     }
 
+    size_t
+    failed(void)
+        const
+    {
+        return m_failed;
+    }
+
+    void
+    finalize(const std::string& reason = "")
+    {
+        if (!m_tcname.empty()) {
+            assert(!reason.empty());
+            got_tc_end(atf::tests::tcr::failed("Bogus test program"));
+        }
+
+        assert(!m_finalized);
+        m_writer.end_tp(reason);
+        m_finalized = true;
+    }
+
     ~muxer(void)
     {
-        m_writer.end_tp();
+        assert(m_finalized);
     }
 };
 
@@ -285,10 +312,15 @@ atf_run::run_test_directory(const atf::fs::path& tp,
                             atf::formats::atf_tps_writer& w)
 {
     atf::atffile af(tp / "Atffile");
-    m_atffile_vars = af.vars();
+    m_atffile_vars = af.conf();
 
     atf::tests::vars_map oldvars = m_config_vars;
-    read_config(af.ts());
+    {
+        atf::tests::vars_map::const_iterator iter =
+            af.props().find("test-suite");
+        assert(iter != af.props().end());
+        read_config((*iter).second);
+    }
 
     bool ok = true;
     for (std::vector< std::string >::const_iterator iter = af.tps().begin();
@@ -404,26 +436,70 @@ atf_run::run_test_program_parent(const atf::fs::path& tp,
 
     // Process the test case's output and multiplex it into our output
     // stream as we read it.
-    muxer(tp, w, resin).read(outin, errin);
+    try {
+        muxer m(tp, w, resin);
+        std::string fmterr;
+        try {
+            m.read(outin, errin);
+        } catch (const atf::parser::parse_errors& e) {
+            fmterr = e.what();
+        } catch (const atf::serial::format_error& e) {
+            fmterr = e.what();
+        }
 
-    outin.close();
-    errin.close();
-    resin.close();
+        outin.close();
+        errin.close();
+        resin.close();
 
-    int status;
-    if (::waitpid(pid, &status, 0) == -1)
-        throw atf::system_error("atf_run::run_test_program(" + tp.str() +
-                                ")", "waitpid(2) failed", errno);
-
-    int code;
-    if (WIFEXITED(status)) {
-        code = WEXITSTATUS(status);
-        //if (failed > 0 && code == EXIT_SUCCESS) {
-            // XXX Bogus test.
-        //}
-    } else
-        code = EXIT_FAILURE;
-    return code;
+        int code, status;
+        if (::waitpid(pid, &status, 0) == -1) {
+            m.finalize("waitpid(2) on the child process " +
+                       atf::text::to_string(pid) + " failed" +
+                       (fmterr.empty() ? "" : (".  " + fmterr)));
+            code = EXIT_FAILURE;
+        } else {
+            if (WIFEXITED(status)) {
+                code = WEXITSTATUS(status);
+                if (m.failed() > 0 && code == EXIT_SUCCESS) {
+                    code = EXIT_FAILURE;
+                    m.finalize("Test program returned success but some test "
+                               "cases failed" +
+                               (fmterr.empty() ? "" : (".  " + fmterr)));
+                } else {
+                    code = fmterr.empty() ? code : EXIT_FAILURE;
+                    m.finalize(fmterr);
+                }
+            } else if (WIFSIGNALED(status)) {
+                code = EXIT_FAILURE;
+                m.finalize("Test program received signal " +
+                           atf::text::to_string(WTERMSIG(status)) +
+                           (WCOREDUMP(status) ? " (core dumped)" : "") +
+                           (fmterr.empty() ? "" : (".  " + fmterr)));
+            } else if (WIFSTOPPED(status)) {
+                code = EXIT_FAILURE;
+                m.finalize("Test program stopped due to signal " +
+                           atf::text::to_string(WSTOPSIG(status)) +
+                           (fmterr.empty() ? "" : (".  " + fmterr)));
+            } else
+                throw std::runtime_error
+                    ("Child process " + atf::text::to_string(pid) +
+                     " terminated with an unknown status condition " +
+                     atf::text::to_string(status));
+        }
+        return code;
+    } catch (...) {
+        // XXX This catch block is here to prevent a SIGPIPE in FreeBSD 6.2
+        // when the construction of the muxer fails.  This should be handled
+        // by the other try block inside this one, but at the moment we do
+        // some reads straight from the constructor which is proving to be
+        // a very annoying approach.
+        outin.close();
+        errin.close();
+        resin.close();
+        int status;
+        (void)::waitpid(pid, &status, 0);
+        throw;
+    }
 }
 
 int
@@ -443,8 +519,20 @@ atf_run::run_test_program(const atf::fs::path& tp,
         assert(false);
         errcode = EXIT_FAILURE;
     } else {
-        errcode = run_test_program_parent(tp, w, outpipe, errpipe, respipe,
-                                          pid);
+        try {
+            errcode = run_test_program_parent(tp, w, outpipe, errpipe,
+                                              respipe, pid);
+        } catch (const atf::parser::parse_errors& e) {
+            // Parse errors captured here were caused by errors in the
+            // child's header, which were not handled by the function.
+            std::string fmterr = e.what();
+            for (std::string::size_type i = 0; i < fmterr.length(); i++)
+                if (fmterr[i] == '\n' || fmterr[i] == '\r')
+                    fmterr[i] = ' ';
+            w.start_tp(tp.str(), 0);
+            w.end_tp(fmterr);
+            errcode = EXIT_FAILURE;
+        }
     }
 
     return errcode;
@@ -506,7 +594,7 @@ int
 atf_run::main(void)
 {
     atf::atffile af(atf::fs::path("Atffile"));
-    m_atffile_vars = af.vars();
+    m_atffile_vars = af.conf();
 
     std::vector< std::string > tps;
     tps = af.tps();
@@ -519,7 +607,12 @@ atf_run::main(void)
     }
 
     // Read configuration data for this test suite.
-    read_config(af.ts());
+    {
+        atf::tests::vars_map::const_iterator iter =
+            af.props().find("test-suite");
+        assert(iter != af.props().end());
+        read_config((*iter).second);
+    }
 
     atf::formats::atf_tps_writer w(std::cout, count_tps(tps));
 
